/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.binding.soap.listener.incoming;

import static org.ow2.petals.binding.soap.Constants.SOAP.FAULT_SERVER;

import java.io.UnsupportedEncodingException;
import java.security.Principal;
import java.util.Iterator;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.InOptionalOut;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.soap.SOAPBody;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axiom.soap.SOAPFactory;
import org.apache.axis2.AxisFault;
import org.apache.axis2.context.MessageContext;
import org.apache.axis2.util.MessageContextBuilder;
import org.apache.ws.security.WSSecurityEngineResult;
import org.apache.ws.security.handler.WSHandlerConstants;
import org.apache.ws.security.handler.WSHandlerResult;
import org.ow2.petals.binding.soap.Constants;
import org.ow2.petals.binding.soap.util.Marshaller;
import org.ow2.petals.binding.soap.util.SUPropertiesHelper;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.MEPType;
import org.ow2.petals.component.framework.message.ExchangeImpl;

/**
 * An abstract request processor. It is used by the external listener to handle
 * incoming requests such as SOAP, REST... Created on 14 févr. 08
 * 
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since
 */
public abstract class RequestProcessor {
	/**
	 * The logger.
	 */
	private Logger logger = null;
	/**
	 * The channel.
	 */
	private DeliveryChannel channel = null;
	/**
	 * The component.
	 */
	private AbstractComponent component = null;
	/**
	 * The exchange factory.
	 */
	private MessageExchangeFactory exchangeFactory = null;

	/**
	 * The constructor.
	 * 
	 * @param component
	 *            the component
	 * @param channel
	 *            the channel
	 * @param log
	 *            the logger
	 */
	public RequestProcessor(final AbstractComponent component, final DeliveryChannel channel, final Logger log) {
		this.component = component;
		this.channel = channel;
		logger = log;
		exchangeFactory = channel.createExchangeFactory();
	}

	protected final void closeMessageExchange(final MessageExchange exchange) throws AxisFault {
		// close the messageExchange with the JBI NMR
		try {
			exchange.setStatus(ExchangeStatus.DONE);
			getChannel().send(exchange);
		} catch (final MessagingException e) {
			final String errorMsg = "Error while closing JBI MessageExchange.";
			getLogger().log(Level.WARNING, errorMsg, e);
			throw new AxisFault(errorMsg, FAULT_SERVER, e);
		}
	}

	/**
	 * Create a message exchange matching a specified mep.
	 * If the MEP is not defined, it will be IN_OPTIONAL_OUT.
	 * 
	 * @param mep the mep
	 * @return a message exchange matching the specified pattern
	 * @throws MessagingException
	 *             if the message exchange can't be created
	 */
	protected final MessageExchange createMessageExchange(final MEPType mep) throws MessagingException {
		MessageExchange exchange = null;
		if (mep != null) {
			if (MEPType.IN_ONLY.equals(mep)) {
				exchange = exchangeFactory.createInOnlyExchange();
			} else if (MEPType.ROBUST_IN_ONLY.equals(mep)) {
				exchange = exchangeFactory.createRobustInOnlyExchange();
			} else if (MEPType.IN_OUT.equals(mep)) {
				exchange = exchangeFactory.createInOutExchange();
			} else if (MEPType.IN_OPTIONAL_OUT.equals(mep)) {
				exchange = exchangeFactory.createInOptionalOutExchange();
			}
		} else {
			exchange = exchangeFactory.createInOptionalOutExchange();
		}
		return exchange;
	}

	/**
	 * Create the JBI message exchange.
	 * 
	 * @param inContext
	 *            the SOAP message
	 * @param operation
	 *            the operation
	 * @param mep
	 *         	  the mep            
	 * @param axisService
	 *            the axis service into petals
	 * @throws AxisFault
	 *             the axis fault
	 * @return the message exchange
	 */
	protected final MessageExchange createMessageExchange(final MessageContext inContext, final QName operation, 
			final MEPType mep, final PetalsAxisService axisService)
			throws AxisFault, MessagingException {

		ExchangeImpl exchangeWrapper = null;
		Source source = null;
		ServiceEndpoint endpoint = null;
		try {
			// create the message exchange from the extensions.
			final MessageExchange exchange = this.createMessageExchange(mep);

		    ConfigurationExtensions extension = axisService.getConsumesCDKExtensions();
		    boolean axis1Compatibility = SUPropertiesHelper.isAxis1CompatibilityEnabled(extension);
		        
			if (inContext.getAttachmentMap() == null || inContext.getAttachmentMap().getContentIDSet() != null
					&& inContext.getAttachmentMap().getContentIDSet().size() <= 0) {
				source = Marshaller.createSourceContent(inContext.getEnvelope(), axis1Compatibility);
			} else {
				// That's avoid to put the attachment in the payload as binary
				// node
				source = Marshaller.createSourceContentAndAttachment(inContext);
			}

			exchangeWrapper = new ExchangeImpl(exchange);
			exchangeWrapper.setOperation(operation);
			// Set the service name, endpoint name and interface name. The
			// Router will find the JBI endpoint to contact with these
			// informations
			exchangeWrapper.setInterfaceName(axisService.getConsumes().getInterfaceName());
			if (axisService.getConsumes().getServiceName() != null) {
				exchangeWrapper.setService(axisService.getConsumes().getServiceName());
				if (axisService.getConsumes().getEndpointName() != null) {
					endpoint = component.getContext().getEndpoint(axisService.getConsumes().getServiceName(), axisService.getConsumes().getEndpointName());
					if (endpoint == null) {
						throw new AxisFault("No endpoint satisfying service-name and endpoint name is found: " + axisService.getConsumes().getServiceName().toString() + ", " + axisService.getConsumes().getEndpointName());
					}
				}
				exchangeWrapper.setEndpoint(endpoint);
			}
			exchangeWrapper.setInMessageContent(source);
			// add SOAP attachments to normalized message
			Marshaller.setAttachments(inContext.getAttachmentMap(), exchangeWrapper.getInMessage());

			// retrieve security subject if rampart module is engaged
			if (isSecuredService(axisService)) {
				final Subject subject = createSecuritySubject(inContext);
				if (subject != null) {
					exchangeWrapper.getInMessage().setSecuritySubject(subject);
				}
			}

			// Get the options from the SOAP message and put them into the JBI
			// message
			Marshaller.setProperties(inContext, exchangeWrapper.getInMessage());

		} catch (final AxisFault e) {
			throw e;
		} catch (final Exception e) {
			throw AxisFault.makeFault(e);
		}

		return exchangeWrapper.getMessageExchange();
	}

	/**
	 * Create the security {@link Subject} from the {@link MessageContext}.<br />
	 * TODO : Check if there are security features in REST too.
	 * 
	 * @param messageContext
	 *            the message context
	 * @return the security subject created from the message context if we have
	 *         found {@link Principal} in the message context else return null;
	 */
	@SuppressWarnings("unchecked")
	protected final Subject createSecuritySubject(final MessageContext messageContext) {
		getLogger().fine("Trying to create JAAS Security Subject");

		Subject subject = null;
		final Vector wssec = (Vector) messageContext.getProperty(WSHandlerConstants.RECV_RESULTS);
		if (wssec != null) {
			for (final Iterator iterator = wssec.iterator(); iterator.hasNext();) {
				final WSHandlerResult wshandler = (WSHandlerResult) iterator.next();
				final Vector wsSecEngineResults = wshandler.getResults();

				for (final Object object : wsSecEngineResults) {
					final WSSecurityEngineResult wsresult = (WSSecurityEngineResult) object;

					final Principal principal = (Principal) wsresult.get(WSSecurityEngineResult.TAG_PRINCIPAL);
					if (principal != null) {
						if (subject == null) {
							subject = new Subject();
						}

						getLogger().fine("Got a principal from input message context : " + principal.getName());
						subject.getPrincipals().add(principal);
					}
				}
			}
		}
		return subject;
	}

	/**
	 * Get the channel.
	 * 
	 * @return the channel
	 */
	public final DeliveryChannel getChannel() {
		return channel;
	}

	/**
	 * Get the logger.
	 * 
	 * @return the logger
	 */
	public final Logger getLogger() {
		return logger;
	}

	/**
	 * Get the processor name. This name MUST be unique ie two processors can
	 * not have the same name !
	 * 
	 * @return the processor name
	 */
	protected abstract String getProcessorName();

	protected final void handleFault(final MessageExchange exchange, final MessageContext outMessage, final SOAPFactory factory) throws AxisFault {
		// exchange FAULT : return a fault message
		if (getLogger().isLoggable(Level.FINE)) {
			getLogger().fine("We found a fault into the JBI exchange.");
		}
		try {
			final SOAPEnvelope envelope = Marshaller.createSOAPEnvelope(factory, exchange.getFault(), true);
			outMessage.setEnvelope(envelope);
		} catch (final AxisFault e) {
			final String errorMsg = "Error while creating SOAP Fault response.";
			getLogger().log(Level.WARNING, errorMsg, e);
			throw new AxisFault(errorMsg, FAULT_SERVER, e);
		}
	}

	/**
	 * Put the {@link NormalizedMessage} attachments in the output Axis2
	 * {@link MessageContext}.
	 * 
	 * @param soapFactory
	 *            the SOAP factory
	 * @param nm
	 *            the normalized message
	 * @param outMessage
	 *            the SOAP message
	 * @throws UnsupportedEncodingException
	 *             the unsupported encoding exception
	 */
	protected final void handleResponseAttachments(final SOAPFactory soapFactory, final NormalizedMessage nm, final MessageContext outMessage)
			throws UnsupportedEncodingException {
		// process the output message with its attachments
		Marshaller.fillSOAPBodyWithAttachments(nm, soapFactory, outMessage);

		final SOAPEnvelope env = outMessage.getEnvelope();
		if (env != null && logger.isLoggable(Level.FINE)) {
			logger.log(Level.FINE, "SOAPENVELOPE AFTER Attachment handling");
			logger.log(Level.FINE, outMessage.getEnvelope().toString());
		}
	}

	/**
	 * Check is the service is secured ie rampart module has been engaged by the
	 * service unit.
	 * 
	 * @param service
	 *            the axis service into petals
	 * @return true if it's securised, false else
	 */
	protected final boolean isSecuredService(final PetalsAxisService service) {
		final ConfigurationExtensions extensions = service.getConsumesCDKExtensions();
		final String modules = extensions.get(Constants.ServiceUnit.MODULES);
		return modules != null && modules.indexOf(Constants.Axis2.RAMPART_MODULE) >= 0;
	}

	/**
	 * Process the request.
	 * 
	 * @param inContext
	 *            the SOAP context
	 * @param soapfactory
	 *            the SOAP factory
	 * @return the out message context
	 * @throws AxisFault
	 *             the axis fault
	 * @throws UnsupportedEncodingException
	 *             the unsupported encoding exception
	 */
	protected abstract MessageContext process(final MessageContext inContext, final SOAPFactory soapfactory) throws AxisFault, UnsupportedEncodingException;

	/**
	 * Process the JBI response. <br>
	 * According to the response status :
	 * <ul>
	 * <li>set outContext to null for DONE</li>
	 * <li>throws an AxisFault for ERROR</li>
	 * <li>set outContext body to a FAULT</li>
	 * <li>set outContext body to a response</li>
	 * </ul>
	 * 
	 * @param exchange
	 *            JBI response
	 * @param inMessage
	 *            the SOAP message
	 * @param factory
	 *            soap factory used to create body response
	 * @return a FAULT or OUT JBI response or null if no response
	 * @throws AxisFault
	 *             with ERROR JBI status, or problem while creating response
	 * @throws UnsupportedEncodingException
	 *             the unsupported encoding exception
	 */
	protected MessageContext processJBIResponse(final MessageExchange exchange, final MessageContext inMessage, final SOAPFactory factory) throws AxisFault,
			UnsupportedEncodingException {
		if (exchange.getStatus().equals(ExchangeStatus.DONE)) {
			// exchange DONE, nothing to do, no soap response
			getLogger().fine("Receive a Done status message");
			return null;

		} else if (exchange.getStatus().equals(ExchangeStatus.ERROR)) {
			// exchange ERROR: throws an axis fault
			getLogger().fine("Receive an Error status message");
			throw new AxisFault(FAULT_SERVER, exchange.getError());

		} else if (exchange.getStatus().equals(ExchangeStatus.ACTIVE)) {
			getLogger().fine("Receive an Active status message");
			final MessageContext outMessage = MessageContextBuilder.createOutMessageContext(inMessage);
			outMessage.getOperationContext().addMessageContext(outMessage);
			try {
				if (exchange.getFault() != null) {
					handleFault(exchange, outMessage, factory);
				} else {
					// Get the output message if the message is an instance of
					// out
					try {
						NormalizedMessage nm = null;
						if (exchange instanceof InOptionalOut || exchange instanceof InOut) {
							nm = exchange.getMessage(Exchange.OUT_MESSAGE_NAME);
						}
						if (nm == null) {
							final String errorMsg = "The MEP '" + exchange.getPattern() + "' does not accept a null response";
							getLogger().log(Level.WARNING, errorMsg);
							throw new AxisFault(errorMsg, FAULT_SERVER);
						} else {
							exchange.getInterfaceName();
							final SOAPEnvelope envelope = Marshaller.createSOAPEnvelope(factory, nm, false);
							outMessage.setEnvelope(envelope);
							handleResponseAttachments(factory, nm, outMessage);
						}
					} catch (final AxisFault e) {
						final String errorMsg = "Error while creating SOAP response";
						getLogger().log(Level.WARNING, errorMsg, e);
						throw new AxisFault(errorMsg, FAULT_SERVER, e);
					}
				}
			} finally {
				closeMessageExchange(exchange);
			}
			return outMessage;
		} else {
			throw new AxisFault(Constants.SOAP.ERROR_WRONG_MESSAGE_STATUS + " " + exchange.getStatus().toString());
		}
	}

	/**
	 * Get the operation from the message context. Assume that the operation is
	 * always the root element of the SOAP body.
	 * 
	 * @param message
	 *            the SOAP message
	 * @return null if the operation has not been found
	 */
	protected final QName retrieveOperationFromSOAPBody(final MessageContext message) {
		QName result = null;
		final SOAPEnvelope env = message.getEnvelope();
		final SOAPBody body = env.getBody();
		final OMElement operation = body.getFirstElement();
		if (operation != null) {
			result = operation.getQName();
		}
		return result;
	}
}
