/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.binding.soap.listener.incoming;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;

import org.ow2.petals.binding.soap.util.NetworkUtil;

import static org.ow2.petals.binding.soap.Constants.HttpServer.DEFAULT_HTTP_ACCEPTORS;
import static org.ow2.petals.binding.soap.Constants.HttpServer.DEFAULT_HTTP_PORT;
import static org.ow2.petals.binding.soap.Constants.HttpServer.DEFAULT_HTTP_THREAD_POOL_SIZE_MAX;
import static org.ow2.petals.binding.soap.Constants.HttpServer.DEFAULT_HTTP_THREAD_POOL_SIZE_MIN;

/**
 * This class is used to store only the SOAP server related configuration
 * values.
 * 
 * @author Christophe HAMERLING - eBMWebSourcing
 * 
 */
public class SoapServerConfig {

	/**
	 * The port used for incoming SOAP requests
	 */
	private int port = DEFAULT_HTTP_PORT;

    /**
     * The host
     */
    private String host;
    
    /**
     * The host IP
     */
    private InetAddress hostIP;

	/**
	 * Restrict access to a particular network interface
	 */
	private boolean restricted = false;

	/**
	 * The servlet mappping : axis2/<b>services</b>
	 */
	private String servicesMapping;

	/**
	 * The servlet context : <b>axis2</b>/services
	 */
	private String servicesContext;

	/**
	 * The protocol used for this server (http / https ...)
	 */
	private String protocol;

	/**
	 * The server provides/!provides the services list
	 */
	private boolean providesList;
	
	private HashMap<String, String> redirects = null;

	/**
	 * Default value is set to 255 (The jetty implementation default)
	 */
	protected int jettyThreadMaxPoolSize = DEFAULT_HTTP_THREAD_POOL_SIZE_MAX;

	/**
	 * Default value is set to 1 (The jetty implementation default)
	 */
	protected int jettyThreadMinPoolSize = DEFAULT_HTTP_THREAD_POOL_SIZE_MIN;

	/**
	 * Default value is set to 4
	 */
	protected int jettyAcceptors = DEFAULT_HTTP_ACCEPTORS;

	protected String wsnContext = "wsn";

	protected String wsnProducerMapping = "producer";

	protected String wsnProducerURL;

	protected String servicesURL;
    
    /**
     * isHostValid
     */
    private boolean isValidHostName;
    
    /**
     * Creates a new instance of {@link SoapServerConfig}
     * 
     */
    public SoapServerConfig() {
        this.restricted = false;
        this.isValidHostName = false;
    }
    
    /**
     * Validate the specified host and add it to the server configuration
     * 
     * @param config
     *            the server configuration
     * 
     * @throws JBIException the host is not valid
     */
    public void setHost(final Logger logger, final String host) {

        Set<Inet4Address> localIPv4Addresses;
        
        try {
            
            localIPv4Addresses = NetworkUtil.getAllLocalIPv4InetAddresses();
        
            if (host == null || host.length() == 0 || host.equals("null")) {
                // no host specified, do not restrict and use all addresses
                try {
                    this.hostIP = getIPv4AddressForHost(localIPv4Addresses);
                    this.isValidHostName = true;
                    this.restricted = false;
                } catch (UnknownHostException uhe) {
                    this.isValidHostName = false;
                    if(logger.isLoggable(Level.WARNING)) {
                        logger.log(Level.WARNING, uhe.getMessage(), uhe);
                    }
                }

            } else {
                // a host is specified, check if is valid
                
                // get the IP address from the host name
                InetAddress address = null;
                try {
                    address = InetAddress.getByName(host);
                } catch (final UnknownHostException uhe) {
                    this.isValidHostName = false;
                    if(logger.isLoggable(Level.WARNING)) {
                        logger.log(Level.WARNING, uhe.getMessage(), uhe);
                    }
                }
    
                // the IP address exists, check if it is a local one
                if (localIPv4Addresses.contains(address)) {
                    this.hostIP = address;
                    this.isValidHostName = true;
                    this.host = host;
                    this.restricted = true;
                } else {
                    this.isValidHostName = false;
                    if(logger.isLoggable(Level.WARNING)) {
                        logger.log(Level.WARNING, host + " does not correspond to a local host");
                    }
                }
            }
        
        } catch (SocketException se) {
            this.isValidHostName = false;
            if(logger.isLoggable(Level.WARNING)) {
                logger.log(Level.WARNING, se.getMessage(), se);
            }
        }
    }
    
    private static final InetAddress getIPv4AddressForHost(Set<Inet4Address> localIPv4Addresses)
            throws UnknownHostException {
        InetAddress ipAddress = null;

        // take the first IP v4 address which is not the loopback one
        InetAddress hostIPv4Address = null;
        if (localIPv4Addresses != null && localIPv4Addresses.size() > 0) {
            final Iterator<Inet4Address> localIPv4AddressesIter = localIPv4Addresses.iterator();
            while (localIPv4AddressesIter.hasNext() && hostIPv4Address == null) {
                final InetAddress ipv4Addressr = localIPv4AddressesIter.next();
                if (!ipv4Addressr.isLoopbackAddress()) {
                    hostIPv4Address = ipv4Addressr;
                }
            }
        }
        
        // if no IP v4 found, use localhost
        if(hostIPv4Address == null) {
            ipAddress = InetAddress.getLocalHost();
        } else {
            ipAddress = hostIPv4Address;
        }
        
        return ipAddress;
    }


	/**
	 * Build the services URL from the current configuration properties
	 * 
	 * @return
	 */
	private String buildServicesURL() {
		final StringBuffer sb = new StringBuffer(getBaseURL());
		sb.append('/');
		sb.append(getServicesContext());
		sb.append('/');
		sb.append(getServicesMapping());
		return sb.toString();
	}

	/**
	 * 
	 * @return
	 */
	public String getBaseURL() {
		if (host == null || host.length() == 0 || host.equals("null")) {
			String baseURL = getProtocol() + "://" + this.getHostAddress()
					+ ":" + port;
			return baseURL;
		} else {
			String baseURL = getProtocol() + "://" + this.host + ":" + port;
			return baseURL;
		}
	}

    /**
     * Return if the host is valid
     * 
     * @return true if the host is valid, otherwise false
     */
    public boolean isValidHostName() {
        return this.isValidHostName;
    }
    
    /**
     * Return if the HTTP/HTTPS server is configured to listen to only one interface or not
     * 
     * @return true if the HTTP/HTTPS server listens to only one interface, otherwise false
     */
    public boolean isRestricted() {
        return this.restricted;
    }
    
    /**
     * Return the HTTP/HTTPS host name
     * 
     * @return the HTTP/HTTPS host name
     */
    public String getHostAddress() {
        return this.hostIP.getHostAddress();
    }
    
    /**
     * Return the HTTP/HTTPS host names (to display)
     * 
     * @return the HTTP/HTTPS host names (to display)
     */
    public String getHostToDisplay() {
        String hostToDisplay;
        
        if(!this.isValidHostName) {
            hostToDisplay = "Invalid Host Name";  
        } else if(this.isRestricted()) {
            String ipAddress = this.hostIP.getHostAddress();
            if(ipAddress.equals(host)) {
                hostToDisplay = "Host : " + ipAddress;
            } else {
                String hostName = this.hostIP.getHostName();
                hostToDisplay = "Host : " + ipAddress + " (" + hostName + ")";
            }
        } else {
            try {
                Set<Inet4Address> localIPv4Addresses = NetworkUtil.getAllLocalIPv4InetAddresses();
    
                Iterator<Inet4Address> localIPv4AddressesIterator = localIPv4Addresses.iterator();
                StringBuffer hostList = new StringBuffer("Hosts : ");
                while(localIPv4AddressesIterator.hasNext()) {
                    Inet4Address localIPv4Address = localIPv4AddressesIterator.next();
                    String hostName = localIPv4Address.getHostName();
                    String ipAddress = localIPv4Address.getHostAddress();
                    if(hostName.equals(ipAddress)) {
                        hostList.append(ipAddress);
                    } else {
                        hostList.append(ipAddress + " (" + hostName + ")");
                    }
                    
                    if(localIPv4AddressesIterator.hasNext()) {
                        hostList.append(" / ");
                    }
                }
                hostToDisplay = hostList.toString();
            } catch (SocketException e) {
                hostToDisplay = "Host : Unable to display the host names";  
                e.printStackTrace();
            }
        }
        
        return hostToDisplay;
    }

	/**
	 * @return the jettyAcceptors
	 */
	public int getJettyAcceptors() {
		return jettyAcceptors;
	}

	/**
	 * @return the jettyThreadMaxPoolSize
	 */
	public int getJettyThreadMaxPoolSize() {
		return jettyThreadMaxPoolSize;
	}

	/**
	 * @return the jettyThreadMinPoolSize
	 */
	public int getJettyThreadMinPoolSize() {
		return jettyThreadMinPoolSize;
	}

	/**
	 * @return the port
	 */
	public int getPort() {
		return port;
	}

	/**
	 * @return the protocol
	 */
	public String getProtocol() {
		return protocol;
	}

	/**
	 * @return the servicesContext
	 */
	public String getServicesContext() {
		return servicesContext;
	}

	/**
	 * @return the servicesMapping
	 */
	public String getServicesMapping() {
		return servicesMapping;
	}

	/**
	 * Get the services URL
	 * 
	 * @return
	 */
	public String getServicesURL() {
		if (servicesURL == null) {
			servicesURL = buildServicesURL();
		}
		return servicesURL;
	}

	/**
	 * @return the providesList
	 */
	public boolean isProvidesList() {
		return providesList;
	}

	/**
	 * @param jettyAcceptors
	 *            the jettyAcceptors to set
	 */
	public void setJettyAcceptors(final int jettyAcceptors) {
		this.jettyAcceptors = jettyAcceptors;
	}

	/**
	 * @param jettyThreadMaxPoolSize
	 *            the jettyThreadMaxPoolSize to set
	 */
	public void setJettyThreadMaxPoolSize(final int jettyThreadMaxPoolSize) {
		this.jettyThreadMaxPoolSize = jettyThreadMaxPoolSize;
	}

	/**
	 * @param jettyThreadMinPoolSize
	 *            the jettyThreadMinPoolSize to set
	 */
	public void setJettyThreadMinPoolSize(final int jettyThreadMinPoolSize) {
		this.jettyThreadMinPoolSize = jettyThreadMinPoolSize;
	}

	/**
	 * @param port
	 *            the port to set
	 */
	public void setPort(final int port) {
		this.port = port;
	}

	/**
	 * @param protocol
	 *            the protocol to set
	 */
	public void setProtocol(final String protocol) {
		this.protocol = protocol;
	}

	/**
	 * @param providesList
	 *            the providesList to set
	 */
	public void setProvidesList(final boolean providesList) {
		this.providesList = providesList;
	}

	/**
	 * @param restrict
	 *            the restrict to set
	 */
	public void setRestrict(final boolean restrict) {
		this.restricted = restrict;
	}

	/**
	 * @param servicesContext
	 *            the servicesContext to set
	 */
	public void setServicesContext(final String servicesContext) {
		this.servicesContext = servicesContext;
	}

	/**
	 * @param servicesMapping
	 *            the servicesMapping to set
	 */
	public void setServicesMapping(final String servicesMapping) {
		this.servicesMapping = servicesMapping;
	}
	
	/**
	 * Define a HTTP redirection (to customize URLs)
	 * @param from Original URI
	 * @param to Destination URI
	 */
	public void addRedirect(String from, String to) {
		if(redirects == null) redirects = new HashMap<String, String>();
		redirects.put(from, to);
	}
	public String getRedirect(String from) {
		if(redirects == null) return null;
		else return redirects.get(from);
	}
	public void removeRedirect(String from) {
		if(redirects != null) redirects.remove(from);
	}
}
